# Author: Jingyi Jessica Li
# Program Completion Date: 09/18/2011
# Modification Date(s): 02/17/2012
# Copyright (c) 2011, Jingyi Jessica Li lijy03@gmail.com
# All rights reserved.
# Under the modified BSD license.

#library(AlgDesign)
library(penalized, verbose=FALSE)
library(nnls, verbose=FALSE)

# build non-overlapping exons
if_overlap <- function(int1, int2) {
  !(int1[2]<=int2[1] | int2[2]<=int1[1])
}

buildExons <- function(raw_exons) {
  if (length(raw_exons) > 2 & length(which(duplicated(raw_exons[,1]) & duplicated(raw_exons[,2]))) >0) {
    raw_exons <- raw_exons[-which(duplicated(raw_exons[,1]) & duplicated(raw_exons[,2])),]
  }
  
  if (length(raw_exons)==2) raw_exons <- matrix(raw_exons, ncol=2)
  
  bds <- sort(unique(raw_exons))
  m <- length(bds) - 1
  intervals <- cbind(bds[1:m], bds[2:(m+1)])
  intervals[1:m, 2] <- intervals[1:m, 2] - c(rep(1,m-1), 0)
  marks <- apply(intervals, 1, FUN=function(int1) {
    any(apply(raw_exons, 1, FUN=function(int2) if_overlap(int1, int2)))
  })
  intervals <- intervals[marks,]
  if (length(intervals)==2) intervals <- matrix(intervals, ncol=2)
  lens <- intervals[,2] - intervals[,1]
  singular_list <- NULL
  min_len <- 1
  if (length(lens) > 1) {
    arg1 <- any(lens < min_len)
    while (arg1) {
      short_idx <- setdiff(which(lens < min_len), singular_list)
      k <- short_idx[1]
      if (k>1 & k<nrow(intervals)) {
        a <- intervals[k+1,1] == intervals[k,2] + 1
        b <- intervals[k,1] == intervals[k-1,2] + 1
        if (a & b) {
          if (lens[k+1] <= lens[k-1]) {
            intervals[k+1,1] <- intervals[k,1]
          } else {
            intervals[k-1,2] <- intervals[k,2]
          } 
          intervals <- intervals[-k,]
        } else if (a) {
          intervals[k+1,1] <- intervals[k,1]
          intervals <- intervals[-k,]
        } else if (b) {
          intervals[k-1,2] <- intervals[k,2]
          intervals <- intervals[-k,]
        } else {
          singular_list <- c(singular_list, k)
        }
      } else if (k==1) {
        if (intervals[k+1,1] == intervals[k,2] + 1) {
          intervals[k+1,1] <- intervals[k,1]
          intervals <- intervals[-k,]
        } else {
          singular_list <- c(singular_list, k)
        }
      } else if (k==nrow(intervals)) {
        if (intervals[k,1] == intervals[k-1,2] + 1) {
          intervals[k-1,2] <- intervals[k,2]
          intervals <- intervals[-k,]
        } else {
          singular_list <- c(singular_list, k)
        }
      }
      
      #if (length(intervals) == 0) {
      #  break
      #  next
      #}
      if (length(intervals)==2) intervals <- matrix(intervals, ncol=2)
      lens <- intervals[,2] - intervals[,1]
      if (length(singular_list) == 0) {
        arg1 <- any(lens < min_len)
      } else {
        arg1 <- any(lens[-singular_list] < min_len)
      }
    }
  }
  exons <- intervals

  # match starting and ending positions of exons to those of raw_exons
  raw_pos_list <- as.numeric(raw_exons)
  for (k in 1:nrow(exons)) {
    for (j in 1:2) {
      temp1 <- raw_pos_list - exons[k,j]
      temp2 <- which(abs(temp1)==1)
      if (length(temp2) > 0) exons[k,j] <- raw_pos_list[temp2[1]]
    }
  }
  return(exons)
}

# associate transcripts with new exons
renameTranscripts <- function(G_is, old_G, new_G) {
	if_cover <- function(int1, int2) {
		int2[1]>=int1[1] & int2[2]<=int1[2]
	}
	old_to_new_idx <- apply(old_G, 1, FUN=function(x) {
		which(apply(new_G, 1, FUN=function(y) if_cover(x,y)))
	})
	n <- nrow(new_G)
	new_G_is <- t( apply( G_is, 1, FUN=function(G_i) {
		new_G_i <- rep(0, n)
		old_exons <- which(G_i == 1)
		new_exons <- unlist(old_to_new_idx[old_exons])
		new_G_i[new_exons] <- 1
		return(new_G_i)
	} ) )
	return(new_G_is)
}

# build paired end bins, e.g. (1,1,2,2)

buildBins <- function(G) {
  n <- nrow(G)
  bins <- cbind(1:n, 1:n, 1:n, 1:n)
  if (n >= 2) {
	nchoose2 <- combn(n,2)
	temp2 <- apply(nchoose2, 2, FUN=function(x) {
		c( c(x[1], x[1], x[1], x[2]), c(x[1], x[1], x[2], x[2]), c(x[1], x[2], x[2], x[2]) )
	})
	bins <- rbind(bins, matrix(temp2, ncol=4, byrow=TRUE))
  }
  if (n >= 3) {
	nchoose3 <- combn(n,3)
	temp3 <- apply(nchoose3, 2, FUN=function(x) {
		c( c(x[1], x[1], x[2], x[3]), c(x[1], x[2], x[2], x[3]), c(x[1], x[2], x[3], x[3]) )
	})
	bins <- rbind(bins, matrix(temp3, ncol=4, byrow=TRUE))
  }
  if (n >= 4) {
	nchoose4 <- combn(n,4)
	temp4 <- apply(nchoose4, 2, FUN=function(x) {
		c( x[1], x[2], x[3], x[4] )
	})
	bins <- rbind(bins, matrix(temp4, ncol=4, byrow=TRUE))
  }
  return(bins)
}

# associate a read to a paired-end bin
read2bin <- function(read, G) {
	temp <- sapply(read, FUN=function(x) which(apply(G, 1, FUN=function(int) x>=int[1] & x<=int[2])))
	if (is.list(temp)) {
		return(0:1)
	} else return (temp)
}

# count reads in paired end bins
countBins <- function(read_table, bins, G) {
  if (length(read_table)==0) {
    return(rep(0,nrow(bins)))
  } else {
    if (length(read_table)==4) read_table <- matrix(read_table, nrow=1)
    mat1 <- matrix(rep(10^(0:3),nrow(bins)), nrow=nrow(bins), byrow=TRUE)
    convert_bins <- rowSums(mat1*bins)
    read_table_convert <- matrix(NA, nrow=nrow(read_table), ncol=ncol(read_table))
    for (i in 1:4) {
      for (j in 1:nrow(G)) {
        s <- G[j,1]
        e <- G[j,2]
        read_table_convert[which(read_table[,i]>=s & read_table[,i]<=e), i] <- j
      }
    }
    rm.ind <- which(is.na(rowSums(read_table)))
    if (length(rm.ind>0)) {
      read_table_convert <- read_table_convert[-rm.ind,]
    }
    if (length(read_table_convert)==0) {
      return(rep(0,nrow(bins)))
    } else {
      if (length(read_table_convert)==4) read_table_convert <- matrix(read_table_convert, nrow=1)
      mat2 <- matrix(rep(10^(0:3),nrow(read_table_convert)), nrow=nrow(read_table_convert), byrow=TRUE)
      convert_counts <- rowSums(mat2*read_table_convert)
      bin_counts <- sapply(convert_bins, FUN=function(x) {
        idx <- which(convert_counts==x)
        length(idx)
      })
      return(bin_counts)
    }
  }
}

# filter out zero-count bins
filterBins <- function(bins, bin_counts) {
  a <- which(bin_counts>0)
  filtered_bins <- bins[a,]
  if (length(a)==1) filtered_bins <- matrix(filtered_bins, nrow=1)
  return(filtered_bins)
}


# filter transcripts with unobserved exon junctions
filterTranscripts <- function(G_is, filtered_bins) {
  junc_mat <- NULL
  for (j in 1:nrow(filtered_bins)) {
    if (filtered_bins[j,1] != filtered_bins[j,2]) junc_mat <- rbind(junc_mat, filtered_bins[j,1:2])
    if (filtered_bins[j,3] != filtered_bins[j,4]) junc_mat <- rbind(junc_mat, filtered_bins[j,3:4])
    if (filtered_bins[j,2] + 1 == filtered_bins[j,3]) {
      junc_mat <- rbind(junc_mat, filtered_bins[j,2:3])
    } else if (filtered_bins[j,2] + 1 < filtered_bins[j,3]) {
      missing_exons <- (filtered_bins[j,2]+1):(filtered_bins[j,3]-1)
      if (sum(sapply(missing_exons, FUN=function(i) any(apply(filtered_bins, 1, FUN=function(x) i %in% x))))==0) {
        junc_mat <- rbind(junc_mat, filtered_bins[j,2:3])
      }
    }    
  }

  single_exons <- which(sapply(1:ncol(G_is), FUN=function(i) {
    temp <- which(apply(filtered_bins, 1, FUN=function(x) i %in% x))
    length(temp) == 1 & length(unique(filtered_bins[temp,])) == 1
  }))

  # handle the case where any two single exons are contiguous
  if (length(single_exons)>1) {
	single_exons_sorted <- sort(single_exons)
	single_exons_diff <- single_exons_sorted[2:length(single_exons_sorted)] - single_exons_sorted[1:(length(single_exons_sorted)-1)]
	for (i in which(single_exons_diff==1)) {
		junc_mat <- rbind(junc_mat, single_exons_sorted[i:(i+1)])
	}
  }

  if (is.null(nrow(junc_mat))) {
    junc_exp <- sum(2^junc_mat)
  } else {
    junc_exp <- rowSums(2^junc_mat)
  }
  a <- duplicated(junc_exp)
  if (length(which(a)) > 0) {
    junc_mat <- junc_mat[-which(a),]
    junc_exp <- junc_exp[-which(a)]
  }
  #print(junc_mat)
  
  
  trpt_idx <- NULL
  for (k in 1:nrow(G_is)) {
    e <- which(G_is[k,] != 0)
    if (length(e) >= 2) {
      tmp <- TRUE
      for (p in 2:length(e)) {
        tmp <- tmp & (sum(2^e[(p-1):p]) %in% junc_exp)
      }
      if (tmp) trpt_idx <- c(trpt_idx, k)
    }
  }
  
  if (length(single_exons)>0) {
    temp1 <- which(rowSums(G_is)==1)
    #print(temp1)
    G_is_temp <- G_is[temp1,]
    for (i in single_exons) {
      #print(G_is_temp)
      if (length(temp1)==1) G_is_temp <- matrix(G_is_temp, nrow=1)
      temp2 <- which(apply(G_is_temp, 1, FUN=function(x) x[i]==1)) 
      trpt_idx <- c(trpt_idx,  temp1[temp2])
    }
  }
  #print(trpt_idx)
  #print(G_is)
  #print(G_is[trpt_idx,])

  

  if (length(trpt_idx)==0) {
    trpt_idx <- 1:nrow(G_is)
  }
  G_is_filtered <- G_is[trpt_idx,]
  if (length(trpt_idx)==1) G_is_filtered <- matrix(G_is_filtered, nrow=1)
  return(G_is_filtered)
}

# build F matrix
F_paired <- function(G_i, G, bin_list, read_len, 
                     lower_len, upper_len, mean, sd, exon_den, dist) 
{ 
  #
  #
  # return a vector for different b_j's
  #
  #
  if (length(dim(bin_list))==0) bin_list <- matrix(bin_list, nrow=1)
  cond_probs <- apply(bin_list, MARGIN=1, FUN=f_paired, G_i=G_i, G=G, read_len=read_len, lower_len=lower_len, upper_len=upper_len, mean=mean, sd=sd, exon_den=exon_den, dist=dist)
  return(as.numeric(t(cond_probs)))
}

f_paired <- function(b_j, G_i, G, read_len, lower_len, upper_len, mean, sd, exon_den, dist) {
  exons <- unique(b_j)
  if (b_j[1]==b_j[2] & b_j[3]==b_j[4]) {
    tmp <- TRUE
  } else if (b_j[1]!=b_j[2] & b_j[3]==b_j[4]) {
    tmp <- sum(G_i[b_j[1]:b_j[2]])==2
  } else if (b_j[1]==b_j[2] & b_j[3]!=b_j[4]) {
    tmp <- sum(G_i[b_j[3]:b_j[4]])==2
  } else {
    tmp <- sum(G_i[b_j[1]:b_j[2]])==2 & sum(G_i[b_j[3]:b_j[4]])==2
  }
  if (all(G_i[exons]==1) & tmp) {
    s <- min(exons)
    e <- max(exons)
    L1 <- G[s,2] - G[s,1]
    if (length(exons) == 1) {
      l1 <- 0
      u1 <- L1 - read_len
      l2 <- read_len
      u2 <- L1
    } else if (length(exons) == 2) {
      L2 <- G[e,2] - G[e,1]
      if ((b_j[2]==b_j[1]) & (b_j[4]==b_j[3])) {
        l1 <- 0
        u1 <- L1 - read_len
        l2 <- L1 + read_len
        u2 <- L1 + L2
        if (e - s > 1) {
          a <- rep(0,length(G_i))
          a[c(s,e)] <- 1
          for (i in which(G_i-a==1)) {
            if (i>s & i<e) {
              l2 <- l2 + G[i,2] - G[i,1]
              u2 <- u2 + G[i,2] - G[i,1]
            }
          }
        }
      } else if ((b_j[2]==b_j[1]) & (b_j[4]>b_j[3])) {
        l1 <- 0
        u1 <- L1 - read_len
        l2 <- L1
        u2 <- L1 + read_len
      } else if ((b_j[2]>b_j[1]) & (b_j[4]==b_j[3])) {
        l1 <- L1 - read_len
        u1 <- L1
        l2 <- L1 + read_len
        u2 <- L1 + L2
      } else if ((b_j[2]>b_j[1]) & (b_j[4]>b_j[3])) {
        l1 <- L1 - read_len
        u1 <- L1
        l2 <- L1
        u2 <- L1 + read_len
      }     
    } else if (length(exons) == 3) {
      m <- exons[2]
      L2 <- G[m,2] - G[m,1]
      L3 <- G[e,2] - G[e,1]
      if (b_j[2]==b_j[1]) {
        l1 <- 0
        u1 <- L1 - read_len
        l2 <- L1 + L2
        u2 <- L1 + L2 + read_len
        if (m-s > 1) {
          a <- rep(0,length(G_i))
          a[c(s,m)] <- 1
          for (i in which(G_i-a==1)) {
            if (i>s & i<m) {
              l2 <- l2 + G[i,2] - G[i,1]
              u2 <- u2 + G[i,2] - G[i,1]
            }
          }
        }
      } else if (b_j[4]==b_j[3]) {
        l1 <- L1 - read_len
        u1 <- L1
        l2 <- L1 + L2 + read_len
        u2 <- L1 + L2 + L3
        if (e-m > 1) {
          a <- rep(0,length(G_i))
          a[c(m,e)] <- 1
          for (i in which(G_i-a==1)) {
            if (i>m & i<e) {
              l2 <- l2 + G[i,2] - G[i,1]
              u2 <- u2 + G[i,2] - G[i,1]
            }
          }
        }
      } else {
        l1 <- L1 - read_len
        u1 <- L1
        l2 <- L1 + L2
        u2 <- L1 + L2 + read_len
      }
    } else if (length(exons)==4) {
      m1 <- exons[2]
      m2 <- exons[3]
      L2 <- G[m1,2] - G[m1,1]
      L3 <- G[m2,2] - G[m2,1]
      L4 <- G[e,2] - G[e,1]
      l1 <- L1 - read_len
      u1 <- L1
      l2 <- L1 + L2 + L3
      u2 <- L1 + L2 + L3 + read_len
      if (m2-m1 > 1) {
        a <- rep(0,length(G_i))
        a[c(m1,m2)] <- 1
        for (i in which(G_i-a==1)) {
          if (i>m1 & i<m2) {
            l2 <- l2 + G[i,2] - G[i,1]
            u2 <- u2 + G[i,2] - G[i,1]
          }
        }
      }
    }
	if (length(l1)<length(lower_len)) l1 <- rep(l1, length(lower_len))
	if (length(u1)<length(lower_len)) u1 <- rep(u1, length(lower_len))
	if (length(l2)<length(lower_len)) l2 <- rep(l2, length(lower_len))
	if (length(u2)<length(lower_len)) u2 <- rep(u2, length(lower_len))
    result <- sapply(1:length(lower_len) , FUN=function(i) {
      Int_Exp(l1[i], u1[i], l2[i], u2[i], lower_len[i], upper_len[i], mean[i], sd[i], exon_den[s], dist)
    })
  } else {
    result <- rep(0, length(lower_len))
  }
  return(result)
}

Int_Exp <- function(l1, u1, l2, u2, lower_len, upper_len, mean, sd, den, dist) {

  if (dist=="trunExp") {
    temp_FUN <- function(x) {
      sapply(x, FUN=function(i) {
        max((pexp(min(u2-i, upper_len),1/mean)-pexp(max(l2-i, lower_len),1/mean))/(pexp(upper_len,1/mean)-pexp(lower_len,1/mean)), 0)
      })
    }
  } else if (dist=="Normal") {
    temp_FUN <- function(x) {
      sapply(x, FUN=function(i) {
        max((pnorm(min(u2-i, upper_len),mean=mean,sd=sd)-pnorm(max(l2-i, lower_len),mean=mean,sd=sd))/(pnorm(upper_len,mean=mean,sd=sd)-pnorm(lower_len,mean=mean,sd=sd)), 0)
      })
    }
  }
  
  integrate(temp_FUN, lower=l1, upper=u1, stop.on.error=FALSE)$val * den
}

build_F <- function(G_is, filtered_bins, G, read_lengths, 
                    frag_param, dist=c("trunExp", "Normal")) {
  if (length(frag_param)==4) frag_param <- matrix(frag_param, nrow=1)
  frag_lower_bd <- frag_param[,1]
  frag_upper_bd <- frag_param[,2]
  frag_mean <- frag_param[,3]
  frag_sd <- frag_param[,4]
  num_exons <- nrow(G)
  exon_lengths <- G[,2]-G[,1]+1
    
  # if GC contents are not included, the density is uniform
  densities <- apply(G_is, 1, FUN=function(G_i) {
      temp <- rep(0, num_exons)
      temp[which(G_i==1)] <- 1/sum(exon_lengths[which(G_i==1)])
      return(temp)
  })
  if (nrow(G_is)==1) densities <- matrix(densities, ncol=1)    
  F <- sapply(1:nrow(G_is), FUN=function(j) {
    F_paired(G_i=G_is[j,], 
             G=G, 
             bin_list=filtered_bins, 
             read_len=read_lengths, 
             lower_len=frag_lower_bd, 
             upper_len=frag_upper_bd, 
             mean=frag_mean, 
             sd=frag_sd, 
             exon_den=densities[,j], 
             dist=dist)
  })
  
  return(F)
}

# discover isoforms and estimate their proportions

learn_lambda <- function(F, bin_counts, bins, G, G_is, lambda=NULL, total_n_reads) {
	# lambda candidates
	lambdas <- c(0.000001, 0.0001, 0.001, seq(0.01, 0.09, by=0.03), seq(0.1, 1, by=0.1))
	
	# the total number of reads for each read length
	total_counts <- rowSums( bin_counts )
	
	# the number of read lengths
	n_read_len <- length(total_counts)
	
	# the number of bins
	n_bins <- nrow(bins)
	
	# list the reads as bin indices, with each element corresponding to a read length
	reads_in_bin_indices <- sapply(1:n_read_len, FUN=function(i) {
		temp <- NULL
		for (j in 1:n_bins) {
			temp <- c( temp, rep(j, bin_counts[i,j]) )
		}
		return(temp)
	}, simplify=FALSE)
	
	# sample a half of the reads in each run, with n_resamp runs in total.  calculate bin counts with the sampled reads
	n_resamp <- 20
	bin_counts_resamp <- sapply(1:n_resamp, FUN=function(k) {
		bin_counts <- t( sapply(1:n_read_len, function(i) {
			temp_reads <- sample( reads_in_bin_indices[[i]], size=ceiling(total_counts[i]/2) )
			temp_bin_counts <- sapply(1:n_bins, FUN=function(j) {
				sum( temp_reads==j )
			})
			return( temp_bin_counts )
		}) ) # each row corresponding to a read length
		return( bin_counts )
	}, simplify=FALSE) 
	
	# normalize the bin counts for isoform discovery / abundance estimation
	b_resamp <- sapply(1:length(n_resamp), FUN=function(i) {
		b <- bin_counts_resamp[[i]]
		b <- b / matrix( rep(rowSums(b),  n_bins), ncol=n_read_len ) * matrix( rep(total_n_reads/mean(total_n_reads), n_read_len), ncol=n_read_len )
		as.numeric( t(b) )
	}) # a matrix with each column for a read length
	bin_counts_normalized_resamp <- sapply(1:length(n_resamp), FUN=function(i) {
		bin_counts <- bin_counts_resamp[[i]]
		bin_counts <- bin_counts / matrix( rep(total_n_reads/mean(total_n_reads), n_read_len), ncol=n_read_len )
		as.numeric( t(b) )
	}) # a matrix with each column for a read length
	
	# if there is only one bin or one transcript, there is no need to use Lasso.  Return lambda as NULL
	if ( length(b) == 1 | nrow(G_is) == 1 ) return(NULL)
	
	# otherwise, run SLIDE for every run.  select lambda by stability.
	stability_results <- rep(NA, length(lambdas))
	G_is_nexons <- rowSums(G_is)
	AW <- F * matrix( rep(G_is_nexons, n_bins*n_read_len), nrow=n_bins*n_read_len, byrow=TRUE ) * matrix( rep(total_n_reads/mean(total_n_reads),length(b)/length(total_n_reads)*ncol(F)), ncol=ncol(F))
}



estimateTranscriptFrequencies <- function(F, bin_counts, bins, G, G_is, lambda=NULL, total_n_reads, mode="discovery") {

  if (nrow(bin_counts)>1) {
    #sum_bin_counts <- colSums(bin_counts)
    #bin_counts <- bin_counts[,which(sum_bin_counts>0)]
	print(bin_counts)
	print(total_n_reads)
    b <- bin_counts / matrix(rep(rowSums(bin_counts),ncol(bin_counts)), ncol=ncol(bin_counts)) *  matrix(rep(total_n_reads/mean(total_n_reads),ncol(bin_counts)), ncol=ncol(bin_counts))
    b <- as.numeric(t(b))
    bin_counts <- bin_counts / matrix(rep(total_n_reads/mean(total_n_reads),ncol(bin_counts)), ncol=ncol(bin_counts))
    bin_counts <- as.numeric(t(bin_counts))
  } else {
    bin_counts <- bin_counts[which(bin_counts>0)]
    b <- bin_counts/sum(bin_counts)
  }

  if (length(b)==1) {
	non_zero_trpts <- matrix(0, nrow=1, ncol=ncol(G_is))
	non_zero_trpts[1,unique(bins)] <- 1
	lens <- sum(G[unique(bins),2]-G[unique(bins),1]+1)
	fit <- bin_counts
  } else {
	
	if (mode=="discovery") {
		if (nrow(G_is)>1) {
			G_is_nexons <- rowSums(G_is)
			AW <- F * matrix(rep(G_is_nexons,length(b)),nrow=length(b),byrow=TRUE) * matrix(rep(total_n_reads/mean(total_n_reads),length(b)/length(total_n_reads)*ncol(F)), ncol=ncol(F))

			# if lambda = NULL, select it by stability criterion

			fit <- try(penalized(response=b, penalized=AW, lambda1=lambda, lambda2=.000001, maxiter=100, unpenalized = ~0, positive=T, model="linear"))
			while (class(fit) == "try-error") {
				rm_idx <- which.min(b)
				b <- b[-rm_idx]
				AW <- AW[-rm_idx,]
				fit <- try(penalized(response=b, penalized=AW, lambda1=lambda, lambda2=.000001, maxiter=100, unpenalized = ~0, positive=T, model="linear"))
				print(fit)
			}
			fit <- coefficients(fit, "penalized")
			MAX_NUM_ITERATIONS = 10
			counter = 0;
			while (counter < MAX_NUM_ITERATIONS && all(fit==0)) {
				lambda <- max(lambda-0.1, lambda/2);
				fit <- try(penalized(response=b, penalized=AW, lambda1=lambda, lambda2=.000001, maxiter=100, unpenalized = ~0, positive=T, model="linear"));
				fit <- coefficients(fit, "penalized");
				counter = counter + 1;
			}
			if( counter == MAX_NUM_ITERATIONS ) {
				print( paste( "Failed to give any non-zero coefficients after ", MAX_NUM_ITERATIONS, "attempts" ) );
				return( list() );
			}
			fitUf <- fit * G_is_nexons
			fitUf <- fitUf/sum(fitUf)
			#discover_idx <- which(fitUf >= 0.1)
			discover_idx <- which(fitUf > 0)
		} else {
			discover_idx <- 1
		}
	} else {
		discover_idx <- 1:nrow(G_is)
	}
	
	non_zero_trpts <- matrix(unlist(G_is[discover_idx,]), nrow=length(discover_idx))
	lens <- sapply(discover_idx, FUN=function(i) {
		temp <- which(G_is[i,]==1)
		sum(G[temp,2]-G[temp,1]+1)
	})
	
	F_new <- F[,discover_idx]

	if (length(discover_idx)==1) F_new <- matrix(F_new, ncol=1)
	fit <- nnls(F_new, as.numeric(bin_counts))
	fit <- coefficients(fit)
  }
  
  fit <- fit/(lens/1000)/(mean(total_n_reads)/1000000)
  output_idx <- which( fit > 0 )
  est_results <- cbind(apply(non_zero_trpts, 1, FUN=function(x) paste("(",paste(which(x==1),collapse=","),")",sep=""))[output_idx], fit[output_idx], round(fit/sum(fit), 2)[output_idx])
  return( est_results )
}

LassoEst <- function( b, AW, ld ) {
  fit <- penalized(response=b, penalized=AW, lambda1=ld, lambda2=.000001, maxiter=100, unpenalized = ~0, positive=T, model="linear", trace=FALSE)
  fit <- coefficients(fit, "penalized")
  fit[ which( fit > 0) ] <- 1
  return( fit )
}
